// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

// Import necessary eswin header files            (public section)
#include "dsp_cb_if.h"
#include "dsp_hw_if.h"
#include "xi_core_api.h"
#include "defines.h"
#include "es_dsp_types.h"
#include "es_kernel_public_defines.h"
#include "es_pingpong_macro.h"
#include "bev_pool.h"

// User defined operator config stored in DRAM offset            (user-defined)
#define DRAM1_CONFIG_OFFSET (104 * 1024)
#define PER_BEV 256
#define INCHANNEL 1
#define OUTCHANNEL 2

//  Declare operator calculation function            (user-defined)
ES_S32 bev_pool_kernel(int32_t *interval_starts, int32_t *interval_lengths, int32_t *ranks_depth,
                       int32_t *ranks_feat, int32_t *ranks_bev, int8_t *depth, int8_t *tran_feat, int8_t *bev,
                       BEV_POOL_CONFIG_S *op_config);

//  The prepare function on the kernel side, reserved as an extension interface
//  (common part, connected to runtime, function name and ARGs cannot be modified)
ES_S32 prepare(es_dsp_buffer_group *params)
{
    //  Retain the interface, currently no implementation is required
    return ES_DSP_SUCCESS;
}

//  The eval function on the kernel side allows users to fill in the required operator implementation
//  (common part, access runtime, function name and ARGs cannot be modified)
ES_S32 eval(es_dsp_buffer_group *params, es_dsp_buffer_group *inputs, es_dsp_buffer_group *outputs)
{
    //  Use TIME-STAMP to get program execution time
    ES_S32 eval_start, eval_stop;
#pragma no_reorder
    TIME_STAMP(eval_start);
#pragma no_reorder

    //  Step 1: Copy the operator config using dma to DRAM
    BEV_POOL_CONFIG_S *config_ddr = (BEV_POOL_CONFIG_S *)(params->buffers[0].addr);
    BEV_POOL_CONFIG_S *config_dram = (BEV_POOL_CONFIG_S *)(get_available_dram1_base() + (ES_U32)DRAM1_CONFIG_OFFSET);
    {
        const uint8_t CHANNEL = 0;

        uint32_t row_sz = sizeof(BEV_POOL_CONFIG_S);
        uint32_t nrows = 1;
        uint32_t ntiles = 1;
        uint32_t src_row_pitch = 1;
        uint32_t dst_row_pitch = 1;
        uint32_t src_tile_pitch = 1;
        uint32_t dst_tile_pitch = 1;

        //  Initiate the dma copy of data from DDR to DRAM
        idma_copy_3d_desc64(CHANNEL, &config_dram, &config_ddr, IDMA_DISABLE_INT, row_sz, nrows, ntiles,
                            src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch);
        //  Waiting for dma to finish copying data
        while (idma_hw_num_outstanding(CHANNEL) > 0) {
        }
    }

    int8_t *depth_ddr_ptr = (int8_t *)inputs->buffers[0].addr;
    int8_t *feat_ddr_ptr = (int8_t *)inputs->buffers[1].addr;
    int32_t *depth_ranks_ddr_ptr = (int32_t *)inputs->buffers[2].addr;
    int32_t *feat_ranks_ddr_ptr = (int32_t *)inputs->buffers[3].addr;
    int32_t *bev_ranks_ddr_ptr = (int32_t *)inputs->buffers[4].addr;
    int32_t *interval_starts_ddr_ptr = (int32_t *)inputs->buffers[5].addr;
    int32_t *interval_lengths_ddr_ptr = (int32_t *)inputs->buffers[6].addr;
    int8_t *bev_ddr_ptr = (int8_t *)outputs->buffers[0].addr;

    // step 6: Start calculating ofmap0 = ifmap0 + ifmap1, where both input and output data are stored on DRAM
    bev_pool_kernel(interval_starts_ddr_ptr, interval_lengths_ddr_ptr, depth_ranks_ddr_ptr, feat_ranks_ddr_ptr,
                    bev_ranks_ddr_ptr, depth_ddr_ptr, feat_ddr_ptr, bev_ddr_ptr, config_dram);

#pragma no_reorder
    TIME_STAMP(eval_stop);
#pragma no_reorder

    return ES_DSP_SUCCESS;
}

static void dma_copy_index(uint8_t ch, uint32_t pdst, uint32_t psrc, int buffer_size)
{
    uint64_t srcw = (uint64_t)psrc;
    uint64_t dstw = (uint64_t)pdst;
    idma_copy_desc64(ch, &dstw, &srcw, buffer_size, IDMA_DISABLE_INT);
}

static int get_cont_count(int *base, int len, int *gap) {
    int cont_count = len == 1 ? 1 : 2;
    *gap = base[1] - base[0];
    for (int j = 2; j < len; j++){
        if(base[j] - base[j-1] != *gap){
            return cont_count;
        } else {
            cont_count++;
        }
    }
    return cont_count;
}

ES_S32 bev_pool_kernel(int32_t *interval_starts, int32_t *interval_lengths, int32_t *ranks_depth,
                       int32_t *ranks_feat, int32_t *ranks_bev, int8_t *depth, int8_t *tran_feat, int8_t *bev,
                       BEV_POOL_CONFIG_S *op_config)
{
    int32_t bev_num = op_config->nbev;
    int32_t C = op_config->c;
    int32_t H = op_config->h;
    int32_t W = op_config->w;
    int32_t N = op_config->n;
    int32_t D = op_config->d;
    int8_t *psum;
    int8_t *feat_dram;

    xb_vecNx16 vec_out_scale_aux = op_config->requant_aux;
    xb_vec2Nx8 vec_depth, vec_feat, vec_bev, vec_bev_raw, vec_output;
    xb_vec2Nx24 vec_mul_out;

    valign va_feat = IVP_ZALIGN();
    valign va_bev = IVP_ZALIGN();

    xb_vec2Nx8 *__restrict feat_vptr;
    xb_vec2Nx8 *__restrict bev_vptr;

    int *interval_start_dram = (int *)get_available_dram0_base();
    int *interval_length_dram = interval_start_dram + PER_BEV;
    int *ranks_depth_dram = interval_length_dram + PER_BEV;

    int out_bev_size = 0;
    int8_t *out_bev_dram = (int8_t *)get_available_dram1_base();
    int32_t total_dram0 = get_available_dram0_size();

    int *next_ranks_depth = ranks_depth;
    int *next_ranks_feat = ranks_feat;
    int *next_ranks_bev = ranks_bev;

    int handle_count = op_config->bev_off;
    int handle_output_size = 0;

    xthal_dcache_region_invalidate(tran_feat, N * H * W * C);
    xthal_dcache_region_invalidate(depth, N * H * W * D);
    //  Start calculating
    for (ES_S32 i = 0; i < bev_num; i += PER_BEV) {
        int rem = PER_BEV;
        if (i + PER_BEV > bev_num) {
            rem = (bev_num-i) % PER_BEV;
        }

        dma_copy_index(INCHANNEL, (uint32_t)interval_start_dram, (uint32_t)(&interval_starts[i]), rem*4);
        dma_copy_index(INCHANNEL, (uint32_t)interval_length_dram, (uint32_t)(&interval_lengths[i]), rem*4);
        while (idma_hw_num_outstanding(INCHANNEL) > 0) {}

        int input_count = interval_start_dram[rem-1] + interval_length_dram[rem-1] - interval_start_dram[0];
        int aligned_off = (input_count + 63) & ~(63);
        int *ranks_feat_dram = (int *)(ranks_depth_dram + aligned_off);
        int *ranks_bev_dram = (int *)(ranks_feat_dram + aligned_off);
        int dram0_remain = total_dram0 - ((uint32_t)ranks_bev_dram + aligned_off - get_available_dram0_base());
        if (dram0_remain < 0) {
            es_printf("overflow:%d\n",input_count);
            break;
        }

        dma_copy_index(INCHANNEL, (uint32_t)ranks_depth_dram, (uint32_t)next_ranks_depth, input_count*4);
        dma_copy_index(INCHANNEL, (uint32_t)ranks_feat_dram, (uint32_t)next_ranks_feat, input_count*4);
        dma_copy_index(INCHANNEL, (uint32_t)ranks_bev_dram, (uint32_t)next_ranks_bev, input_count*4);
        while (idma_hw_num_outstanding(INCHANNEL) > 0) {}

        next_ranks_depth = next_ranks_depth + input_count;
        next_ranks_feat = next_ranks_feat + input_count;
        next_ranks_bev = next_ranks_bev + input_count;

        out_bev_size = (ranks_bev_dram[input_count-1] - ranks_bev_dram[0] + 1) * C;
        int dram1_remain = DRAM1_CONFIG_OFFSET - out_bev_size;
        if (dram1_remain < 0) {
            es_printf("output overflow:%d\n",out_bev_size);
            for(int m = 0;m<input_count;m++){
                es_printf("index:%d, value:%d\n",handle_count+m, ranks_bev_dram[m]);
            }
            break;
        }

        memset(out_bev_dram, 0, out_bev_size);
        handle_output_size = ranks_bev_dram[0] * C;

        for (int k = 0; k<rem; k++) {
            int start = interval_start_dram[k] - handle_count;
            int len = interval_length_dram[k];
            int off = (ranks_bev_dram[start] - ranks_bev_dram[0]) * C;
            psum = out_bev_dram + off;
            vec_bev_raw = 0;

            int pitch=0;
            int feat_off = ranks_feat_dram[start];
            uint32_t nrows = get_cont_count(&ranks_feat_dram[start], len, &pitch);
            feat_dram = out_bev_dram + out_bev_size;

            if (dram1_remain < len * C) {
                es_printf("no enghout for feat:%d ,dram0 remain:%d dram1 remain:%d\n", len * C, dram0_remain, dram1_remain);
            }

            if (nrows > 1) {
                int8_t *pdst = feat_dram;
                int8_t *psrc = tran_feat + feat_off * C;
                pitch = pitch*C;
                idma_copy_3d_desc64(0, &pdst, &psrc, IDMA_DISABLE_INT, C, nrows, 1, pitch, C, pitch*nrows, C*nrows);
                if(len-nrows>0) {
                    for (int j = nrows; j < len; j++)
                    {
                        int8_t *cur_feat  = tran_feat + ranks_feat_dram[start + j] * C;
                        memcpy((void*)(feat_dram+(j*C)), cur_feat, C);
                    }
                }
                while (idma_hw_num_outstanding(0) > 0) {}

            } else {
                for (int j = 0; j < len; j++) {
                    int8_t *cur_feat  = tran_feat + ranks_feat_dram[start + j] * C;
                    memcpy((void*)(feat_dram+(j*C)), cur_feat, C);
                }
            }

            feat_vptr = (xb_vec2Nx8 *)(feat_dram);
            va_feat = IVP_LA2NX8_PP(feat_vptr);

            for (int j = 0; j < len; j++) {
                int8_t *cur_depth = depth + ranks_depth_dram[start + j];

                IVP_LA2NX8_IP(vec_feat, va_feat, feat_vptr);
                vec_depth = cur_depth[0];

                vec_mul_out = IVP_MUL2NX8(vec_feat, vec_depth);

                xb_vecNx16 m_outEven = IVP_PACKL2NX24_0(vec_mul_out);
                xb_vecNx16 m_outOdd = IVP_PACKL2NX24_1(vec_mul_out);
                xb_vecNx48 m_wvec = IVP_MULNX16(vec_out_scale_aux, m_outEven);
                m_outEven = IVP_PACKVRNX48(m_wvec, op_config->requant_shift);
                m_wvec = IVP_MULNX16(vec_out_scale_aux, m_outOdd);
                m_outOdd = IVP_PACKVRNX48(m_wvec, op_config->requant_shift);
                m_outEven = IVP_MAXNX16(IVP_MINNX16(m_outEven, 127), -128);
                m_outOdd = IVP_MAXNX16(IVP_MINNX16(m_outOdd, 127), -128);
                vec_output = IVP_SEL2NX8I(IVP_MOV2NX8_FROMNX16(m_outOdd), IVP_MOV2NX8_FROMNX16(m_outEven),
                                        IVP_SELI_8B_INTERLEAVE_1_EVEN);

                vec_bev = IVP_ADD2NX8(vec_bev_raw, vec_output);
                vec_bev_raw = vec_bev;
            }
            bev_vptr = (xb_vec2Nx8 *)(psum);
            va_bev = IVP_LA2NX8_PP(bev_vptr);
            IVP_SA2NX8_IP(vec_bev, va_bev, bev_vptr);
            IVP_SAPOS2NX8_FP(va_bev, bev_vptr);
        }

        dma_copy_index(OUTCHANNEL, (uint32_t)(bev + handle_output_size), (uint32_t)out_bev_dram, out_bev_size);
        while (idma_hw_num_outstanding(OUTCHANNEL) > 0) {}

        handle_count += input_count;
    }

    return 0;
}

//  Return pointers to the prepare and eval functions to the framework, non modifiable            (public section)
__attribute__((visibility("default"))) void *get_prepare(void) { return (void *)prepare; }
__attribute__((visibility("default"))) void *get_eval(void) { return (void *)eval; }

