// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include "dsp_cb_if.h"
#include "es_pingpong_macro.h"
#include <math.h>
#include "dma_tile_scheduler.h"
#include "es_kernel_public_defines.h"

uint32_t take_remainder(uint32_t full_size, uint32_t tile_size, int32_t max_cnt, int32_t cur_cnt)
{
    if (0 == full_size % tile_size) {
        return tile_size;
    } else {
        if (cur_cnt == max_cnt - 1) {
            return full_size % tile_size;
        } else {
            return tile_size;
        }
    }
}

void get_offset(dma_param_t* dma_param)
{
    uint32_t row_pitch = 0;
    uint32_t tile_pitch = 0;
    if (dma_param->tile_on_src_flag) {
        row_pitch = dma_param->src_row_pitch;
        tile_pitch = dma_param->src_tile_pitch;
    } else {
        row_pitch = dma_param->dst_row_pitch;
        tile_pitch = dma_param->dst_tile_pitch;
    }

    dma_param->offset_h = dma_param->tile_shape.horizontals * dma_param->loop_cnt_h;
    dma_param->offset_v = dma_param->tile_shape.verticals * dma_param->loop_cnt_v * row_pitch;
    dma_param->offset_c = dma_param->tile_shape.tiles * dma_param->loop_cnt_c * tile_pitch;

    if (dma_param->tile_on_src_flag) {
        dma_param->src_data +=
            (dma_param->offset_h + dma_param->offset_v + dma_param->offset_c) * dma_param->pixel_bytes;
    } else {
        dma_param->dst_data +=
            (dma_param->offset_h + dma_param->offset_v + dma_param->offset_c) * dma_param->pixel_bytes;
    }
}

void set_dma_param(dma_param_t* dma_param, char* src, char* dst, uint32_t row_sz, uint32_t nrows, uint32_t ntiles,
                   int32_t tile_on_src, dma_shape_t tile_shape, dma_shape_t full_shape, uint32_t loop_time_c,
                   uint32_t loop_time_h, uint32_t loop_time_v, uint32_t loop_cnt_c, uint32_t loop_cnt_h,
                   uint32_t loop_cnt_v, uint32_t src_row_pitch, uint32_t dst_row_pitch, uint32_t src_tile_pitch,
                   uint32_t dst_tile_pitch, uint32_t pixel_bytes)
{
    dma_param->row_sz = row_sz;
    dma_param->nrows = nrows;
    dma_param->ntiles = ntiles;

    dma_param->dst_data = dst;
    dma_param->src_data = src;
    dma_param->tile_on_src_flag = tile_on_src;
    dma_param->tile_shape = tile_shape;
    dma_param->full_shape = full_shape;

    dma_param->loop_time_c = loop_time_c;
    dma_param->loop_time_h = loop_time_h;
    dma_param->loop_time_v = loop_time_v;
    dma_param->loop_cnt_c = loop_cnt_c;
    dma_param->loop_cnt_h = loop_cnt_h;
    dma_param->loop_cnt_v = loop_cnt_v;

    dma_param->src_row_pitch = src_row_pitch;
    dma_param->dst_row_pitch = dst_row_pitch;
    dma_param->src_tile_pitch = src_tile_pitch;
    dma_param->dst_tile_pitch = dst_tile_pitch;
    dma_param->pixel_bytes = pixel_bytes;

    get_offset(dma_param);
}

uint32_t get_dma_params(char* src_ping, char* src_pong, char* dst_ping, char* dst_pong, dma_shape_t tile_shape,
                        dma_shape_t full_shape, dma_param_t* dma_param, uint32_t src_row_pitch, uint32_t src_tile_pitch,
                        uint32_t dst_row_pitch, uint32_t dst_tile_pitch, int32_t tile_on_src, uint32_t pixel_bytes)
{
    uint32_t loop_time_h = ceilf((float)full_shape.horizontals / (float)tile_shape.horizontals);
    uint32_t loop_time_v = ceilf((float)full_shape.verticals / (float)tile_shape.verticals);
    uint32_t loop_time_c = ceilf((float)full_shape.tiles / (float)tile_shape.tiles);
    uint32_t loop_time_full = loop_time_h * loop_time_v * loop_time_c;
    uint32_t loop_time_cnt = 0;

    for (int32_t loop_cnt_c = 0; loop_cnt_c < loop_time_c; ++loop_cnt_c) {
        for (int32_t loop_cnt_v = 0; loop_cnt_v < loop_time_v; ++loop_cnt_v) {
            for (int32_t loop_cnt_h = 0; loop_cnt_h < loop_time_h; ++loop_cnt_h) {
                uint32_t row_sz =
                    take_remainder(full_shape.horizontals, tile_shape.horizontals, loop_time_h, loop_cnt_h);
                uint32_t nrows = take_remainder(full_shape.verticals, tile_shape.verticals, loop_time_v, loop_cnt_v);
                uint32_t ntiles = take_remainder(full_shape.tiles, tile_shape.tiles, loop_time_c, loop_cnt_c);

                int32_t dma_param_idx = loop_cnt_h + loop_cnt_v * loop_time_h + loop_cnt_c * loop_time_v * loop_time_h;

                char* dst = loop_time_cnt % 2 == 0 ? dst_ping : dst_pong;
                char* src = loop_time_cnt % 2 == 0 ? src_ping : src_pong;

                set_dma_param(&(dma_param[dma_param_idx]), src, dst, row_sz, nrows, ntiles, tile_on_src, tile_shape,
                              full_shape, loop_time_c, loop_time_h, loop_time_v, loop_cnt_c, loop_cnt_h, loop_cnt_v,
                              src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch, pixel_bytes);

                loop_time_cnt += 1;
            }
        }
    }
    return loop_time_full;
}
void schedule_dma_nidx(dma_param_t* dma_param, int32_t chn)
{
    dma_param_t* cur_param = dma_param;
    uint64_t dstw = (uint64_t)(uint32_t)cur_param->dst_data;
    uint64_t srcw = (uint64_t)(uint32_t)cur_param->src_data;
    ES_KERNEL_LOG("dstw addr: %p - %p srcw addr: %p - %p\n", (char*)dstw, cur_param->dst_data, (char*)srcw,
                  cur_param->src_data);

    uint32_t pixel_size = cur_param->pixel_bytes;
    ES_KERNEL_LOG("cur_param->row_sz:%d , cur_param->nrows:%d, cur_param->ntiles:%d,\n"
                  "cur_param->src_row_pitch:%d , cur_param->dst_row_pitch:%d ,\n"
                  "cur_param->src_tile_pitch:%d , cur_param->dst_tile_pitch:%d pixel_size:%d \n",
                  cur_param->row_sz, cur_param->nrows, cur_param->ntiles, cur_param->src_row_pitch,
                  cur_param->dst_row_pitch, cur_param->src_tile_pitch, cur_param->dst_tile_pitch, pixel_size);

    ES_KERNEL_LOG("to idma_copy_3d_desc64 ch: %d\n", chn);
    int32_t ret =
        idma_copy_3d_desc64(chn, &dstw, &srcw, 0, cur_param->row_sz * pixel_size, cur_param->nrows, cur_param->ntiles,
                            cur_param->src_row_pitch * pixel_size, cur_param->dst_row_pitch * pixel_size,
                            cur_param->src_tile_pitch * pixel_size, cur_param->dst_tile_pitch * pixel_size);
    ES_KERNEL_LOG("idma_copy_3d_desc64 scheduled ch:%d ret:%d \n", chn, ret);
}
void schedule_dma(dma_param_t* dma_param, int32_t idx, int32_t chn)
{
    ES_KERNEL_LOG("schedule_dma entered \n", NULL);
    dma_param_t* cur_param = &(dma_param[idx]);
    ES_KERNEL_LOG("cur_param addr: %p idx:%d \n", cur_param, idx);

    schedule_dma_nidx(cur_param, chn);
}

char* get_tile_inout_addr(int32_t input_flag, uint32_t (*get_base_fn)())
{
    ES_KERNEL_LOG("to get local memory \n", NULL);
    char* tile_input = (char*)get_base_fn();
    ES_KERNEL_LOG("local memory get done addr: %p \n", tile_input);
    if (input_flag) {
        return tile_input;
    } else {
        return tile_input + TILE_SIZE;
    }
}

void tile_dma_copy(DMA_PARAMS_S cur_param, char* src_data, char* dst_data, int32_t chn)
{
    uint64_t srcw = (uint64_t)(uint32_t)(src_data + cur_param.src_offset);
    uint64_t dstw = (uint64_t)(uint32_t)(dst_data + cur_param.dst_offset);

    ES_KERNEL_LOG("to idma_copy_3d_desc64 ch: %d\n", chn);
    int32_t ret = idma_copy_3d_desc64(chn, &dstw, &srcw, 0, cur_param.row_buf_sz, cur_param.nrows, cur_param.tiles,
                                      cur_param.src_row_buf_pitch, cur_param.dst_row_buf_pitch,
                                      cur_param.src_tile_buf_pitch, cur_param.dst_tile_buf_pitch);
    ES_KERNEL_LOG("idma_copy_3d_desc64 scheduled ch:%d ret:%d \n", chn, ret);
}
