// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#pragma once

/**
 * @brief This macro function select the odd and even elements from vec2Nx8s vector,
 *        and compose two vceNx16s vector.
 *
 * @param   vecIn    The input vector register, data type is vec2Nx8s.
 * @param   vecOut0  The output vector register,data type is vecNx16s
 *                   and composed by the even elements of vecIn.
 * @param   vecOut1  The output vector register,data type is vecNx16s
 *                   and composed by the odd elements of vecIn.
 */
#define SEL_ODD_EVEN_2NX8(vecIn, vecOut0, vecOut1) \
    {                                              \
        vecOut0 = IVP_UNPKS2NX8_0(vecIn);          \
        vecOut1 = IVP_UNPKS2NX8_1(vecIn);          \
    }

/**
 * @brief This macro function select the high and low elements from vecNxf16 vector,
 *        and compose two vceN_2xf32 vector.
 *
 * @param   vecIn    The input vector register, data type is vecNxf16.
 * @param   vecOutH  The output vector register,data type is vceN_2xf32
 *                   and composed by the high elements of vecIn.
 * @param   vecOutL  The output vector register,data type is vceN_2xf32
 *                   and composed by the low elements of vecIn.
 */
#define SEL_HIGH_LOW_FROM_F162F32(vecIn, vecOutH, vecOutL)                          \
    {                                                                               \
        xb_vecN_2xf32 vecIn0 = IVP_CVTF32NXF16_0(vecIn);                            \
        xb_vecN_2xf32 vecIn1 = IVP_CVTF32NXF16_1(vecIn);                            \
        IVP_DSELN_2XF32I(vecOutH, vecOutL, vecIn1, vecIn0, IVP_DSELI_INTERLEAVE_2); \
    }

/**
 * @brief This macro function select the high and low elements from vecNx8S vector,
 *        and compose two vceN_2xf32 vector.
 *
 * @param   vecIn    The input vector register, data type is vecNx8S.
 * @param   vecOutH  The output vector register,data type is vceN_2xf32
 *                   and composed by the high elements of vecIn.
 * @param   vecOutL  The output vector register,data type is vceN_2xf32
 *                   and composed by the low elements of vecIn.
 */
#define CVT_NXS8_TO_N2XF32(vecIn, vecOutH, vecOutL)      \
    {                                                    \
        xb_vecNx16 a = xb_vecNx8_rtor_xb_vecNx16(vecIn); \
        xb_vecNx48 c = IVP_MULNX16(a, 1);                \
        xb_vecN_2x32v h = IVP_CVT32SNX48H(c);            \
        xb_vecN_2x32v l = IVP_CVT32SNX48L(c);            \
        vecOutH = xb_vecN_2x32v_rtor_xb_vecN_2xf32(h);   \
        vecOutL = xb_vecN_2x32v_rtor_xb_vecN_2xf32(l);   \
    }

/**
 * @brief Accumulate the partial elements of vceN_2xf32 vector by a mask vector.
 *
 * @param   vecIn    The input vector register, data type is vecNx8S.
 * @param   mask     The mask vector register,data type is vboolN_2,if the lane in the mask
 *                   is true,then the corresponding lane in vecIn will add to sum.
 * @param   sum      The output result,data type is float.
 */
#define RADDN_2XF32T(vecIn, sum, mask)              \
    {                                               \
        xb_vecN_2xf32 vecMulTmp = 0;                \
        IVP_MULN_2XF32T(vecMulTmp, vecIn, 1, mask); \
        sum = IVP_RADDN_2XF32(vecMulTmp);           \
    }

/**
 * @brief Calculate the maximum value in an array.
 *
 * @param   src         The input array, data type is int16.
 * @param   width       The length of input array.
 * @param   maxVal      The max value of input array.
 */
static void esReductionMaxS16(xb_vecNx16* src, int16_t* maxVal, int32_t width)
{
    xb_vecNx16 vecMaxVal = SHRT_MIN;

    xb_vecNx16* restrict pvecSrc = src;

    int x = 0;
    for (; x < width - 3 * XCHAL_IVPN_SIMD_WIDTH; x += 4 * XCHAL_IVPN_SIMD_WIDTH) {
        vboolN vbMask = IVP_LTUNX16(IVP_SEQNX16U(), width - x - 3 * XCHAL_IVPN_SIMD_WIDTH);

        valign vaSrc = IVP_LANX16_PP(pvecSrc);
        xb_vecNx16 vecSrc0;
        xb_vecNx16 vecSrc1;
        xb_vecNx16 vecSrc2;
        xb_vecNx16 vecSrc3;

        IVP_LANX16_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LANX16_IP(vecSrc1, vaSrc, pvecSrc);
        IVP_LANX16_IP(vecSrc2, vaSrc, pvecSrc);
        IVP_LAVNX16_XP(vecSrc3, vaSrc, pvecSrc, 2 * (width - x - 3 * XCHAL_IVPN_SIMD_WIDTH));

        vecMaxVal = IVP_MAXNX16(IVP_MAXNX16(vecMaxVal, vecSrc2), IVP_MAXNX16(vecSrc1, vecSrc0));

        IVP_MAXNX16T(vecMaxVal, vecMaxVal, vecSrc3, vbMask);
    }
    if (x < width - 2 * XCHAL_IVPN_SIMD_WIDTH) {
        vboolN vbMask = IVP_LTUNX16(IVP_SEQNX16U(), width - x - 2 * XCHAL_IVPN_SIMD_WIDTH);

        valign vaSrc = IVP_LANX16_PP(pvecSrc);
        xb_vecNx16 vecSrc0;
        xb_vecNx16 vecSrc1;
        xb_vecNx16 vecSrc2;

        IVP_LANX16_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LANX16_IP(vecSrc1, vaSrc, pvecSrc);
        IVP_LAVNX16_XP(vecSrc2, vaSrc, pvecSrc, 2 * (width - x - 2 * XCHAL_IVPN_SIMD_WIDTH));

        vecMaxVal = IVP_MAXNX16(IVP_MAXNX16(vecSrc0, vecSrc1), vecMaxVal);

        IVP_MAXNX16T(vecMaxVal, vecMaxVal, vecSrc2, vbMask);
    } else if (x < width - XCHAL_IVPN_SIMD_WIDTH) {
        vboolN vbMask = IVP_LTUNX16(IVP_SEQNX16U(), width - x - XCHAL_IVPN_SIMD_WIDTH);

        valign vaSrc = IVP_LANX16_PP(pvecSrc);
        xb_vecNx16 vecSrc0;
        xb_vecNx16 vecSrc1;

        IVP_LANX16_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LAVNX16_XP(vecSrc1, vaSrc, pvecSrc, 2 * (width - x - XCHAL_IVPN_SIMD_WIDTH));

        vecMaxVal = IVP_MAXNX16(vecSrc0, vecMaxVal);

        IVP_MAXNX16T(vecMaxVal, vecMaxVal, vecSrc1, vbMask);
    } else if (x < width) {
        vboolN vbMask = IVP_LTUNX16(IVP_SEQNX16U(), width - x);

        valign vaSrc = IVP_LANX16_PP(pvecSrc);
        xb_vecNx16 vecSrc0;

        IVP_LAVNX16_XP(vecSrc0, vaSrc, pvecSrc, 2 * (width - x));

        IVP_MAXNX16T(vecMaxVal, vecSrc0, vecMaxVal, vbMask);
    }

    *maxVal = IVP_RMAXNX16(vecMaxVal);
}

/**
 * @brief Calculate the maximum value in an array.
 *
 * @param   src         The input array, data type is int8.
 * @param   width       The length of input array.
 * @param   maxVal      The max value of input array.
 */
static void esReductionMaxS8(xb_vec2Nx8* src, int8_t* maxVal, int32_t width)
{
    xb_vec2Nx8 vecMaxVal = SCHAR_MIN;
    xb_vec2Nx8 tmpVecMaxVal = SCHAR_MIN;

    xb_vec2Nx8* restrict pvecSrc = src;
    int8_t maxVal_temp = SCHAR_MIN;
    int x = 0;
    vbool2N vbMask;
    int trailW;
    for (; x < width - 6 * XCHAL_IVPN_SIMD_WIDTH; x += 8 * XCHAL_IVPN_SIMD_WIDTH) {
        trailW = (width - x - 6 * XCHAL_IVPN_SIMD_WIDTH) > 64 ? 64 : (width - x - 6 * XCHAL_IVPN_SIMD_WIDTH);
        vbMask = IVP_LTU2NX8(IVP_SEQ2NX8U(), trailW);
        valign vaSrc = IVP_LA2NX8_PP(pvecSrc);
        xb_vec2Nx8 vecSrc0;
        xb_vec2Nx8 vecSrc1;
        xb_vec2Nx8 vecSrc2;
        xb_vec2Nx8 vecSrc3;

        IVP_LA2NX8_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LA2NX8_IP(vecSrc1, vaSrc, pvecSrc);
        IVP_LA2NX8_IP(vecSrc2, vaSrc, pvecSrc);
        IVP_LAV2NX8_XP(vecSrc3, vaSrc, pvecSrc, width - x - 6 * XCHAL_IVPN_SIMD_WIDTH);
        vecSrc3 = IVP_MOV2NX8T(vecSrc3, tmpVecMaxVal, vbMask);

        vecMaxVal = IVP_MAX2NX8(IVP_MAX2NX8(vecSrc0, vecSrc1), vecMaxVal);
        vecMaxVal = IVP_MAX2NX8(vecMaxVal, vecSrc2);
        vecMaxVal = IVP_MAX2NX8(vecMaxVal, vecSrc3);
    }
    if (x < width - 4 * XCHAL_IVPN_SIMD_WIDTH) {
        trailW = width - x - 4 * XCHAL_IVPN_SIMD_WIDTH;
        vbMask = IVP_LTU2NX8(IVP_SEQ2NX8U(), trailW);
        valign vaSrc = IVP_LA2NX8_PP(pvecSrc);
        xb_vec2Nx8 vecSrc0;
        xb_vec2Nx8 vecSrc1;
        xb_vec2Nx8 vecSrc2;

        IVP_LA2NX8_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LA2NX8_IP(vecSrc1, vaSrc, pvecSrc);
        IVP_LAV2NX8_XP(vecSrc2, vaSrc, pvecSrc, trailW);
        vecSrc2 = IVP_MOV2NX8T(vecSrc2, tmpVecMaxVal, vbMask);

        vecMaxVal = IVP_MAX2NX8(IVP_MAX2NX8(vecSrc0, vecSrc1), vecMaxVal);
        vecMaxVal = IVP_MAX2NX8(vecMaxVal, vecSrc2);
    } else if (x < width - 2 * XCHAL_IVPN_SIMD_WIDTH) {
        trailW = width - x - 2 * XCHAL_IVPN_SIMD_WIDTH;
        vbMask = IVP_LTU2NX8(IVP_SEQ2NX8U(), trailW);
        valign vaSrc = IVP_LA2NX8_PP(pvecSrc);
        xb_vec2Nx8 vecSrc0;
        xb_vec2Nx8 vecSrc1;

        IVP_LA2NX8_IP(vecSrc0, vaSrc, pvecSrc);
        IVP_LAV2NX8_XP(vecSrc1, vaSrc, pvecSrc, trailW);
        vecSrc1 = IVP_MOV2NX8T(vecSrc1, tmpVecMaxVal, vbMask);

        vecMaxVal = IVP_MAX2NX8(IVP_MAX2NX8(vecSrc0, vecSrc1), vecMaxVal);
    } else if (x < width) {
        trailW = width - x;
        vbMask = IVP_LTU2NX8(IVP_SEQ2NX8U(), trailW);
        valign vaSrc = IVP_LA2NX8_PP(pvecSrc);
        xb_vec2Nx8 vecSrc0;

        IVP_LAV2NX8_XP(vecSrc0, vaSrc, pvecSrc, trailW);

        vecSrc0 = IVP_MOV2NX8T(vecSrc0, tmpVecMaxVal, vbMask);
        vecMaxVal = IVP_MAX2NX8(vecMaxVal, vecSrc0);
    }

    *maxVal = IVP_RMAX2NX8(vecMaxVal);
}
