// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.
#pragma once

#include <stdint.h>
#include <xtensa/tie/xt_ivpn.h>

#if defined(COMPILER_XTENSA)
#define PLACE_IN_SRAM __attribute__((section(".sram.data")))
#else
#define PLACE_IN_SRAM
#endif

// #define ES_EXP_F16_F16

#define ARR_SIZE_expfminmax 2
#define ARR_SIZE_expftblf 7

typedef float float32_t;

#define EXPF_ERRH 0

#define __Pragma(a) _Pragma(a)

#ifndef IVP_N
#define IVP_N 32 /* SIMD width in 16-bit elements */
#endif

#if IVP_N == 32
#define IVP_2N (IVP_N * 2)
#define IVP_4N (IVP_N * 4)
#define IVP_N_2 (IVP_N / 2)
#define IVP_N_4 (IVP_N / 4)
/* log2(IVP_N) */
#define LOG2_IVP_N 5
#define LOG2_IVP_4N (LOG2_IVP_N + 2)
#define LOG2_IVP_2N (LOG2_IVP_N + 1)
#define LOG2_IVP_N_2 (LOG2_IVP_N - 1)
#define LOG2_IVP_N_4 (LOG2_IVP_N - 2)
#else
#error unsupported IVP_N
#endif

union ufloat32uint32 {
    uint32_t u;
    float32_t f;
};

#ifdef ES_EXP_F16_F16
#define ES_VECTOR_TYPE xb_vecN_2xf16
#define ES_SCALAR_TYPE xb_f16
#define ES_PIXEL_BYTE 2

#define ES_LOAD_XP(vecIn, xa, pX, width) IVP_LAVN_2XF16_XP(vecIn, xa, pX, width);

#define ES_LOAD_PP(pX) IVP_LAN_2XF16_PP(pX);

#define ES_SAVE_XP(vecOut, za, pZ, width) IVP_SAVN_2XF16_XP(vecOut, za, pZ, width);

#define ES_SAVE_FP(za, pZ) IVP_SAPOSN_2XF16_FP(za, pZ);
#else

#define ES_VECTOR_TYPE xb_vecN_2xf32
#define ES_SCALAR_TYPE float
#define ES_PIXEL_BYTE 4
#define ES_LOAD_XP(vecIn, xa, pX, width) IVP_LAVN_2XF32_XP(vecIn, xa, pX, width);

#define ES_LOAD_PP(pX) IVP_LAN_2XF32_PP(pX);

#define ES_SAVE_XP(vecOut, za, pZ, width) IVP_SAVN_2XF32_XP(vecOut, za, pZ, width);

#define ES_SAVE_FP(za, pZ) IVP_SAPOSN_2XF32_FP(za, pZ);
#endif

/**
 * @brief Vectorized Floating-Point/Half-Precision Floating-Point Exponential
 *        The exponential (or anti-logarithm) function computes the exponential
 *        value e to the power of input vector x[N], and stores the result to output
 *        vector z[N].
 *        Note: The input and output data types need to be the same.
 *
 * @param   z[N]       The output array,support fp16/fp32 data type.
 * @param   x[N]       The input array,support fp16/fp32 data type.
 * @param   N          The length of input array.
 */
static void es_vec_exp(ES_SCALAR_TYPE* z, const ES_SCALAR_TYPE* x, int N)
{
    const union ufloat32uint32 expfminmax[2] = /* minimum and maximum arguments of expf() input */
        {
            {0xc2ce8ed0}, /*-1.0327893066e+002f */
            {0x42b17218}  /* 8.8722839355e+001f */
        };

    const int32_t invln2_Q30 = 1549082005L;             /* 1/ln(2), Q30 */
    const union ufloat32uint32 plusInff = {0x7f800000}; /* +Inf */
    const union ufloat32uint32 expftblf[] = {{0x39655635}, {0x3aa24c7a}, {0x3c1eb2d1}, {0x3d633ddb},
                                             {0x3e75ff24}, {0x3f317212}, {0x3f800000}};

    ES_VECTOR_TYPE vecIn, vecOut;  // diff  xb_vecN_2xf32 xin, xin2, txin, zout;
    xb_vecN_2xf32 xin, xin2, txin, zout;
    xb_vecN_2xf32 p0, p1, p2, p3, p4, p5, p6;
    xb_vecN_2xf32 scl1, scl2;
    xb_vecN_2x32v t, exp_fract, exp_int, e1, e2;
    xb_vecN_2x64w W;
    xb_vecNx16 invln2;
    valign xa, za;
    vboolN_2 b_nan, b_max, b_inf;

    int n;
    const ES_VECTOR_TYPE* __restrict pX = (const ES_VECTOR_TYPE*)x;
    ES_VECTOR_TYPE* __restrict pZ = (ES_VECTOR_TYPE*)z;
    xtfloat* __restrict ptbl = (xtfloat*)expftblf;

    /* common argument checks */
    if (N <= 0) return;

    /* load 1/ln(2) constant in Q30 */
    invln2 = IVP_MOVNX16_FROMN_2X32(IVP_LSN_2X32_I((const int*)&invln2_Q30, 0));

    za = IVP_ZALIGN();
    xa = ES_LOAD_PP(pX);
    // diff   xa = IVP_LAN_2XF32_PP(pX);
    //  xa = IVP_LAN_2XF16_PP(pX);
    __Pragma("loop_count min=1");
    for (n = 0; n < (N + IVP_N_2 - 1) >> LOG2_IVP_N_2; n++) {
        ES_LOAD_XP(vecIn, xa, pX,
                   (uint8_t*)x + N * ES_PIXEL_BYTE -
                       (uint8_t*)pX);  // diff IVP_LAVN_2XF32_XP( xin, xa, pX, (uint8_t*)x + N*4 - (uint8_t*)pX );
        xin = vecIn;

        txin = xin;
        /* Check input for values that are out of domain/range */
        b_nan = IVP_UNN_2XF32(xin, xin);              /* x==NaN                */
        b_max = IVP_ULEN_2XF32(expfminmax[1].f, xin); /* x>=88.72284 or x==NaN */
        b_inf = IVP_UEQN_2XF32(xin, plusInff.f);      /* x==+Inf or x==NaN     */

        /* Limit input values to [-128;+127] and replace NaNs *
         * with some numbers to avoid unnecessary exceptions  */
        xin = IVP_MOVN_2XF32T(127.0f, xin, b_max);
        xin = IVP_MAXN_2XF32(xin, -128.0f);
        /* Convert the input to Q24 and scale to 1/ln(2) */
        t = IVP_TRUNCN_2XF32(xin, 24);
        W = IVP_MULN_2X16X32_0(invln2, t);
        IVP_MULAHN_2X16X32_1(W, invln2, t); /* Q24*Q30->Q54 */

        /* Separate input to positive fractional part and integer part */
        exp_fract = IVP_PACKVRNRN_2X64W(W, 22);    /* Q54->Q32 */
        exp_fract = IVP_SRLIN_2X32U(exp_fract, 1); /* Q32->Q31 */
        exp_int = IVP_PACKHN_2X64W(W);             /* Q54->Q22 */
        exp_int = IVP_SRLIN_2X32(exp_int, 22);     /* Q22->Q0 */

        /* compute 2^fract in floating-point format */
        xin = IVP_FLOATN_2X32(exp_fract, 31); /* scale fraction part by 2^-31 with conversion from int32 to float */

        /* Pass input to the output if x==+INF or x==NaN */
        xin = IVP_MOVN_2XF32T(txin, xin, b_inf);
        /* load polynomial coefficients */
        IVP_LSRN_2XF32_IP(p0, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_IP(p1, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_IP(p2, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_IP(p3, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_IP(p4, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_IP(p5, ptbl, sizeof(float32_t));
        IVP_LSRN_2XF32_XP(p6, ptbl, -6 * (int)sizeof(float32_t));
        /* compute polynomial using combination *
         * of Estrin`s and Horner schemes       */
        xin2 = IVP_MULN_2XF32(xin, xin);  // xin2 = xin*xin;
        IVP_MULAN_2XF32(p1, xin, p0);
        IVP_MULAN_2XF32(p3, xin, p2);
        IVP_MULAN_2XF32(p5, xin, p4);

        IVP_MULAN_2XF32(p3, xin2, p1);
        IVP_MULAN_2XF32(p5, xin2, p3);

        IVP_MULAN_2XF32(p6, xin, p5);

        /* Apply integer exponential part to the result */
        exp_int = IVP_ADDN_2X32(exp_int, 254);
        e1 = IVP_SRLIN_2X32(exp_int, 1);
        e2 = IVP_SUBN_2X32(exp_int, e1);
        e1 = IVP_SLLIN_2X32(e1, 23);
        e2 = IVP_SLLIN_2X32(e2, 23);
        scl1 = IVP_MOVN_2XF32_FROMN_2X32(e1);
        scl2 = IVP_MOVN_2XF32_FROMN_2X32(e2);
        zout = IVP_MULN_2XF32(p6, scl1);
        zout = IVP_MULN_2XF32(zout, scl2);

        vecOut = zout;
        ES_SAVE_XP(vecOut, za, pZ,
                   (uint8_t*)z + N * ES_PIXEL_BYTE -
                       (uint8_t*)pZ);  // diff IVP_SAVN_2XF32_XP( zout, za, pZ, (uint8_t*)z + N*4 - (uint8_t*)pZ );
    }
    ES_SAVE_FP(za, pZ);  // diff  IVP_SAPOSN_2XF32_FP(za, pZ);
} /* esVecExp() */
