/*
 * Copyright (c) 2012-2018 Cadence Design Systems, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* This file implements init and cleanup functions for the library.
   Note that if you enable dynamic expansion of the function table
   then malloc/free will be called and parts of the C library will
   have to be linked in. To avoid this, the default option is to
   not use dynamic expansion. You may have to resize the table to
   suit the needs of your library.
   You can either edit this file to suit your needs and rebuild the
   library, or copy it into your application project and then edit
   it as needed.
 */


#include <stdlib.h>


/* Default size of function table, allow upto 32 atexit functions */
#define TABLE_SIZE      32
/* Whether to use dynamic expansion - default OFF */
#define USE_DYNAMIC     0

/* Typedef for table entry */
typedef void (*funcp)(void *);

/* Function table */
#if USE_DYNAMIC
static funcp * ptable     = NULL;
static int     table_size = 0;
#else
static funcp   exit_table[TABLE_SIZE];
static funcp * ptable     = &exit_table[0];
static int     table_size = TABLE_SIZE;
#endif
static int   index = 0;


/* Register function to be called before library unload */
int
__cxa_atexit (void (*func) (void *), void * arg, void * dso_handle)
{
#if USE_DYNAMIC
    if ((ptable == NULL) || (index >= table_size)) {
        void * newmem;

        table_size += TABLE_SIZE;
        newmem = malloc (table_size * sizeof(funcp));
        if (newmem == NULL) {
            return -1;
        }
        if (ptable != NULL) {
            memcpy (newmem, ptable, sizeof(funcp) * (table_size - TABLE_SIZE));
            free (ptable);
        }
        ptable = newmem;
    }
#endif

    if (index >= table_size) {
        return -1;
    }

    ptable[index] = func;
    index++;
    return 0;
}


/* Handle functions registered via atexit() */
int
atexit (void (*func)(void))
{
    return __cxa_atexit ((funcp) func, NULL, NULL);
}


/* Call all registered atexit functions */
void
__cxa_finalize (void * d)
{
    while (index > 0) {
        --index;
        (*ptable[index]) (NULL);
    }
#if USE_DYNAMIC
    free (ptable);
#endif
}

/* Register __cxa_finalize() to be called automatically 
   when _fini() is called.
 */
__asm__ (
    "\t.section .fini\n"
#ifdef __XTENSA_CALL0_ABI__
    "\tcall0   __cxa_finalize\n"
#else
    "\tcall8   __cxa_finalize\n"
#endif
    "\t.section .text\n"
        );

