/*
 * Copyright (c) 2019-2022 by Cadence Design Systems. ALL RIGHTS RESERVED.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <xtensa/config/core.h>
#include <xt_mld_loader.h>
#include "interface.h"

XTMLD_LIBRARY_NAME(mult_section_lib);

/*
 * Here we define the preference (requirements) for the loading the code
 * sections : .text. .1.text, .2.text, and .3.text
 *
 * Code with minimal performance impact ==> xtmld_load_normal
 * Performance sensitive code           ==> xtmld_load_prefer_localmem
 * Requirement for local mem            ==> xtmld_load_require_localmem
 */
XTMLD_CODE_MEM_REQ(xtmld_load_prefer_localmem, xtmld_load_prefer_localmem,
        xtmld_load_normal, xtmld_load_normal);

/*
 * Here we define the preference (requirements) for the loading the data
 * sections : .data, .1.data, .2.data, and .3.data
 *
 * Data with minimal performance impact ==> xtmld_load_normal
 * Performance sensitive data           ==> xtmld_load_prefer_localmem
 * Requirement for local mem            ==> xtmld_load_require_localmem
 */
XTMLD_DATA_MEM_REQ(xtmld_load_normal, xtmld_load_prefer_localmem,
        xtmld_load_prefer_localmem, xtmld_load_normal);



static struct lib_symbol_tray lib_symbol_tray;
static struct app_symbol_tray* app_tray;

/*
 * Rather than including stdlib.h, we just define the one function that is
 * available and we want: atexit()
 */
int atexit(void (*fn)(void));

void * _start(struct app_symbol_tray* ex);


/* Here we define static and extern variables and place them in
 * different sections. Later we print them and their addresses
 * so we can confirm that they have correct
 * value and are placed in the appropriate data section.
 */

#define BSS_LEN 10000
/* Uninitialized (BSS) data goes in the .data output section */
static int dbss[BSS_LEN];

/* .0.data is goes in .data */
__attribute__ ((section (".0.data"))) static int b_0[] ={ 1, 2, 3, 4, 5, 6, 7 };

__attribute__ ((section (".0.data"))) static int d_0 = 5;

/* default goes into .data section */
static int e_0[] = { 1, 2, 3, 4, 22 };

/* explicitly put this one in .data */
__attribute__ ((section (".data"))) static int f_0 = 5;

/* .1.data */
__attribute__ ((section (".1.data"))) int b_1[] = { 1, 2, 3, 5, 7, 11 };

/* .2.data OR .dram0.data */
#if (defined XCHAL_HAVE_DATARAM0) && XCHAL_HAVE_DATARAM0
__attribute__ ((section (".dram0.data"))) static int b_2[] = { 1, 4, 9, 16, 25 };
#else
__attribute__ ((section (".2.data"))) static int b_2[] = { 1, 4, 9, 16, 25 };
#endif

/* .3.data OR .dram1.data */
#if (defined XCHAL_HAVE_DATARAM1) && XCHAL_HAVE_DATARAM1
__attribute__ ((section (".dram1.data"))) int b_3[] =
{ 1, 8, 27, 64, 125 };
#else
__attribute__ ((section (".3.data"))) int b_3[] =
{ 1, 8, 27, 64, 125 };
#endif


/* Here we define functions and place them in different sections. Later we
 * print their addresses to ensure that they are placed in the correct sections.
 */

/* .0.text goes into .text  */
__attribute__((section (".0.text"))) static void my_other_exit()
{
    app_tray->printf("other atexit() function successfully called\n");
}

void defaults_to_text()
{
    app_tray->printf("functions default to .text\n");
}

__attribute__((section (".text"))) void explicit_to_text()
{
    app_tray->printf("explicit to .text\n");
}

/* .1.text */
__attribute__((section (".1.text"))) static void my_exit()
{
    app_tray->printf("atexit() function successfully called\n");
}

__attribute__((section (".1.text"))) static double foo2(int x)
{
    return 3.0 * x;
}

/* .2.text */
/* This function is called from the main application. */

__attribute__((section (".2.text"))) static const char * foo(void)
{
#if (defined XCHAL_HAVE_DATARAM0) && XCHAL_HAVE_DATARAM0
    int have_dram0 = 1;
#else
    int have_dram0 = 0;
#endif
#if (defined XCHAL_HAVE_DATARAM1) && XCHAL_HAVE_DATARAM1
    int have_dram1 = 1;
#else
    int have_dram1 = 0;
#endif

    /* register atexit() functions */
    atexit(my_exit);
    atexit(my_other_exit);
    app_tray->printf("\n.data\n");
    app_tray->printf("b_0[4] (should be 5) = %d, &b_0[0] (should be in .data) = %p\n",
            b_0[4], &b_0[0]);
    app_tray->printf("d_0 = %d, &d_0 = %p\n", d_0, &d_0);
    app_tray->printf("e_0[4] = %d, &e_0[0] = %p\n", e_0[4], &e_0[0]);
    app_tray->printf("f_0 = %d, &f_0 = %p\n", f_0, &f_0);
    app_tray->printf("dbss[100] = %d, &dbss[0] = %p\n", dbss[100], &dbss[0]);
    app_tray->printf("\n");

    app_tray->printf(".1.data\n");
    app_tray->printf("b_1[4] (should be 7) = %d, &b_1[0] (should be in .1.data) = %p\n",
            b_1[4], &b_1[0]);
    app_tray->printf("\n");

    app_tray->printf("%s\n", have_dram0 ? ".dram0.data" : ".2.data");
    app_tray->printf("b_2[4] (should be 25) = %d, &b_2[0] (should be in %s) = %p\n",
            b_2[4], have_dram0 ? ".dram0.data" : ".2.data", &b_2[0]);
    app_tray->printf("\n");
    app_tray->printf("%s\n", have_dram1 ? ".dram1.data" : ".3.data");
    app_tray->printf("b_3[4] (should be 125) = %d, &b_0[0] (should be in %s) = %p\n",
            b_3[4], have_dram1 ? ".dram1.data" : ".3.data", &b_3[0]);
    app_tray->printf("\n");

    app_tray->printf(".text\n");
    app_tray->printf("&my_other_exit() = %p\n", my_other_exit);
    app_tray->printf("&defaults_to_text() = %p\n", defaults_to_text);
    app_tray->printf("&explicit_to_text() = %p\n", explicit_to_text);
    app_tray->printf(".1.text\n");
    app_tray->printf("&my_exit() = %p\n", my_exit);
    app_tray->printf(".2.text\n");
    app_tray->printf("&foo() = %p\n", foo);
    app_tray->printf(".3.text\n");
    app_tray->printf("&_start() = %p\n", _start);
    explicit_to_text();
    defaults_to_text();
    return "this string returned from foo";
}

/* .3.text */
__attribute__((section(".3.text"))) void * _start(struct app_symbol_tray* ex)
{
    app_tray = ex;
    lib_symbol_tray.foo = foo;
    lib_symbol_tray.foo2 = foo2;
    return &lib_symbol_tray;
}

