/*
 * Copyright (c) 2019-2022 by Cadence Design Systems. ALL RIGHTS RESERVED.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <xtensa/config/core.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "xt_mld_loader.h"

/* Interface between main program and c++ library */
#include "interface.h"

/* This library will be provided by the object created by the
 package_loadable_lib.pl tool.

 Observe that the name declared here must match the name used with
 the "-e" option to the tool.
 */

#ifdef LOAD_LIB_OTF
int lib_buf[20000];
xtmld_packaged_lib_t * p_mult_section_lib;

/* This library is a C++ library and it has similar requirement as previous
 one.  Please use xt-xc++ or xt-clang++ to build this C++ library */
xtmld_packaged_lib_t * p_mult_section_cpp_lib;

#else
extern xtmld_packaged_lib_t mult_section_lib;
xtmld_packaged_lib_t * p_mult_section_lib = &mult_section_lib;

/* This library is a C++ library and it has similar requirement as previous
 one.  Please use xt-xc++ or xt-clang++ to build this C++ library */
extern xtmld_packaged_lib_t mult_section_cpp_lib;
xtmld_packaged_lib_t * p_mult_section_cpp_lib = &mult_section_cpp_lib;
#endif

/* Declare a function pointer of the same type as the _start function
 provided in the overlay. This function can have whatever prototype you
 choose, but for convenience, it most often takes and returns function
 pointers for calling into and out of the loadable library.
 */

typedef struct lib_symbol_tray* (*mult_section_load_start_ptr_t)(
        struct app_symbol_tray*);

mult_section_load_start_ptr_t mult_section_load_start;

/* Declare a function pointer of the same type as the _start function
 provided in the C++ LIBRARY. This function can have whatever prototype you
 choose, but for convenience, it most often takes and returns function
 pointers for calling into and out of the loadable library.
 */

typedef struct cpp_lib_symbol_tray* (*mult_section_cpp_load_start_ptr_t)(
        struct app_symbol_tray*);

mult_section_cpp_load_start_ptr_t mult_section_cpp_load_start;

/* If local datarams exist, reserve some space for loading required
 sections from the dynamic lib.
 */
#if XCHAL_HAVE_DATARAM0
__attribute__ ((section (".dram0.data"))) char dram0_buf[1024];
#endif

#if XCHAL_HAVE_DATARAM1
__attribute__ ((section (".dram1.data"))) char dram1_buf[1024];
#endif

static const char* mem_require_string(xtmld_mem_require_t req)
{
    switch (req)
    {
    case xtmld_load_normal:
        return "normal";
    case xtmld_load_prefer_localmem:
        return "prefer local memory";
    case xtmld_load_require_localmem:
        return "require local memory";
    case xtmld_load_require_dram0:
        return "require dataram0";
    case xtmld_load_require_dram1:
        return "require dataram1";
    case xtmld_load_prefer_l2mem:
        return "prefer L2 memory";
    default:
        return "???";
        break;
    }
}

const char* section_name(const char* name, int i)
{
    static char buffer[10];
    if (i)
        sprintf(buffer, ".%1d.%s", i, name);
    else
        sprintf(buffer, ".%s", name);
    return buffer;
}

const char * data_section_name(int i, xtmld_mem_require_t req)
{
    if (req == xtmld_load_require_dram0)
    {
        return ".dram0.data";
    }
    else if (req == xtmld_load_require_dram1)
    {
        return ".dram1.data";
    }

    return section_name("data", i);
}

#ifdef LOAD_LIB_OTF
static int load_file(char* filename, void* buf, size_t buf_size)
{
    FILE * fp;
    int size;
    int rev = 0;

    if ((fp = fopen(filename, "rb")))
    {
        fseek (fp , 0 ,SEEK_END);
        size = ftell(fp);
        rewind(fp);
    }
    else
    {
        printf("Can't open file %s\n", filename);
        return 0;
    }

    if (size <= buf_size)
    {
        rev = fread(buf, 1, size, fp);
        if (rev != size)
        {
            printf("Loading file %s failed\n", filename);
            return 0;
        }
    }
    else
    {
        printf("Buffer is too small to load the file %s\n", filename);
        return 0;
    }

    // OK
    fclose(fp);
    return 1;
}
#endif

int main(int argc, char * argv[])
{
    xtmld_library_info_t info;
    xtmld_result_code_t code;
    xtmld_state_t loader_state;
    void * data_memory[XTMLD_MAX_DATA_SECTIONS];
    void * code_memory[XTMLD_MAX_CODE_SECTIONS];
    struct lib_symbol_tray* lib_tray;
    struct cpplib_symbol_tray* cpplib_tray;
    struct app_symbol_tray app_tray;

    int i;
    /* export functions to C library */
    app_tray.printf = printf;
    app_tray.abort = abort;
    app_tray.malloc = malloc;
    app_tray.free = free;;
#ifdef LOAD_LIB_OTF
    /* load the packed image on the fly */
    if (!load_file("stripped_mult_section.lib", lib_buf, sizeof(lib_buf)))
    {
        printf("Load library failed\n");
        return -1;
    }

    p_mult_section_lib = (xtmld_packaged_lib_t *)lib_buf;
#endif

    if ((code = xtmld_library_info(p_mult_section_lib, &info)) == xtmld_success)
        printf("xtmld_library_info() succeeded ...\n");
    else
    {
        printf("xtmld_library_info() failed with error: %s\n",
                xtmld_error_string(code));
        return -1;
    }

    /*
     * Allocate memory for the sections using malloc().  This assumes
     * that the heap memory will be executable ... if this is not the
     * case a different allocator must be used. For sections that must
     * be in local dataram, use the static buffers.
     *
     * In a more complete implementation the section memory requirement
     * should be checked and the appropriate type of memory should
     * be used.
     */
    printf(" section   %8s    %8s  load preference\n", "size", "address");
    for (i = 0; i < info.number_of_code_sections; i++)
    {
        if (info.code_section_size[i])
            code_memory[i] = malloc(info.code_section_size[i]);
        else
            code_memory[i] = 0;
        printf("%8s 0x%08x  0x%08x  %s\n", section_name("text", i),
                info.code_section_size[i], (unsigned) code_memory[i],
                mem_require_string(info.code_section_mem_req[i]));
    }
    printf("\n");
    for (i = 0; i < info.number_of_data_sections; i++)
    {
        if (info.data_section_size[i])
        {
#if XCHAL_HAVE_DATARAM0
            if (info.data_section_mem_req[i] == xtmld_load_require_dram0)
            data_memory[i] = dram0_buf;
#if XCHAL_HAVE_DATARAM1
            else if (info.data_section_mem_req[i] == xtmld_load_require_dram1)
            data_memory[i] = dram1_buf;
#endif
            else
#endif
            data_memory[i] = malloc(info.data_section_size[i]);
        }
        else
            data_memory[i] = 0;
        printf("%11s 0x%08x  0x%08x  %s\n",
                data_section_name(i, info.data_section_mem_req[i]),
                info.data_section_size[i], (unsigned) data_memory[i],
                mem_require_string(info.data_section_mem_req[i]));
    }
    printf("\n\n");

    if (((code = xtmld_load(p_mult_section_lib, code_memory, data_memory,
            &loader_state, (xtmld_start_fn_ptr*) &mult_section_load_start))
            == xtmld_success) && (mult_section_load_start != NULL))
        printf("xtmld_load() succeeded, start_fn = %p\n",
                mult_section_load_start);
    else
    {
        printf("xtmld_load() failed: %s\n", xtmld_error_string(code));
        return -1;
    }

    /* We successfully loaded the library. Now call it's entry point so it can
     get a pointer to printf, and it will return a pointer to its foo
     function.

     If we wanted to use more functions, we could pass and return arrays of
     pointers.

     Also, if the interface to the library is just a single function, then
     the entry point itself could serve as that function.
     */

    const char * string_returned_from_foo;

    lib_tray = mult_section_load_start(&app_tray);
    string_returned_from_foo = lib_tray->foo();
    printf(" function returned the string \"%s\"\n", string_returned_from_foo);
    printf("foo2(8) => %lf\n", lib_tray->foo2(8));

    /* unloading library */
    printf("about to unload library...\n");
    printf("Observe that atexit functions are called.\n");

    code = xtmld_unload((xtmld_state_t*) &loader_state);

    if (code == xtmld_success)
        printf("xtmld_unload() succeeded.\n");
    else
        printf("xtmld_unload failed: %s\n", xtmld_error_string(code));

    for (i = 0; i < info.number_of_code_sections; i++)
        free(code_memory[i]);
    for (i = 0; i < info.number_of_data_sections; i++)
    {
        if ((info.data_section_mem_req[i] != xtmld_load_require_dram0)
                && (info.data_section_mem_req[i] != xtmld_load_require_dram1))
        {
            free(data_memory[i]);
        }
    }

    printf("\n\n");

#ifdef LOAD_LIB_OTF
    /* load the packed image on the fly */
    if (!load_file("stripped_mult_section_cpp.lib", lib_buf, sizeof(lib_buf)))
    {
        printf("Load library failed\n");
        return -1;
    }

    p_mult_section_cpp_lib = (xtmld_packaged_lib_t *)lib_buf;
#endif

    /* 
     Load and run the C++ library, this C++ library uses exception handling.
     So, it needs some supporting functions from this main module.  Please
     see interface.h for detail
     */
    if ((code = xtmld_library_info(p_mult_section_cpp_lib, &info))
            == xtmld_success)
        printf("C++ library: xtmld_library_info() succeeded ...\n");
    else
    {
        printf("C++ library: xtmld_library_info() failed with error: %s\n",
                xtmld_error_string(code));
        return -1;
    }

    /*
     * Allocate memory for the sections using malloc().  This assumes
     * that the heap memory will be executable ... if this is not the
     * case a different allocator must be used.
     *
     * In a more complete implementation the section memory requirement
     * should be checked and the appropriate type of memory should
     * be used.
     */
    printf(" section   %8s    %8s  load preference\n", "size", "address");
    for (i = 0; i < info.number_of_code_sections; i++)
    {
        if (info.code_section_size[i])
            code_memory[i] = malloc(info.code_section_size[i]);
        else
            code_memory[i] = 0;
        printf("%8s 0x%08x  0x%08x  %s\n", section_name("text", i),
                info.code_section_size[i], (unsigned) code_memory[i],
                mem_require_string(info.code_section_mem_req[i]));
    }
    printf("\n");
    for (i = 0; i < info.number_of_data_sections; i++)
    {
        if (info.data_section_size[i])
            data_memory[i] = malloc(info.data_section_size[i]);
        else
            data_memory[i] = 0;
        printf("%11s 0x%08x  0x%08x  %s\n",
                data_section_name(i, info.data_section_mem_req[i]),
                info.data_section_size[i], (unsigned) data_memory[i],
                mem_require_string(info.data_section_mem_req[i]));
    }
    printf("\n\n");

    if (((code = xtmld_load(p_mult_section_cpp_lib, code_memory, data_memory,
            &loader_state, (xtmld_start_fn_ptr*) &mult_section_cpp_load_start))
            == xtmld_success) && (mult_section_cpp_load_start != NULL))
        printf("C++ library: xtmld_load() succeeded, start_fn = %p\n",
                mult_section_cpp_load_start);
    else
    {
        printf("C++ library: xtmld_load() failed: %s\n",
                xtmld_error_string(code));
        return -1;
    }

    /* We successfully loaded the library. Now call it's entry point with a pointer to
     array of functions that it needs, and it will return a pointer to its foo function.

     If we wanted to use more functions, we could pass and return arrays of
     pointers.

     Also, if the interface to the library is just a single function, then
     the entry point itself could serve as that function.
     */

    cpplib_tray =
            (struct cpplib_symbol_tray*) mult_section_cpp_load_start(
                    &app_tray);
    string_returned_from_foo = cpplib_tray->foo();
    printf("C++ library: function returned the string \"%s\"\n",
            string_returned_from_foo);

    printf("otherfoo(xyz, 5) => %lf\n", cpplib_tray->otherfoo("xyz", 5));

    /* unloading library */
    printf("about to unload C++ library...\n");
    printf("Observe that atexit functions are called.\n");

    code = xtmld_unload((xtmld_state_t*) &loader_state);

    if (code == xtmld_success)
        printf("xtmld_unload() C++ library succeeded\n.");
    else
        printf("xtmld_unload C++ library failed: %s\n",
                xtmld_error_string(code));

    for (i = 0; i < info.number_of_code_sections; i++)
        free(code_memory[i]);
    for (i = 0; i < info.number_of_data_sections; i++)
        free(data_memory[i]);
    return 0;
}

