// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

// Import necessary eswin header files            (public section)
#include "dsp_cb_if.h"
#include "dsp_hw_if.h"
#include "xi_core_api.h"
#include "defines.h"
#include "es_dsp_types.h"
#include "es_kernel_public_defines.h"
#include "es_pingpong_macro.h"
#include "user_add.h"

// User defined operator config stored in DRAM offset            (user-defined)
#define DRAM1_CONFIG_OFFSET (104 * 1024)

//  Declare operator calculation function            (user-defined)
ES_S32 add_kernel(int8_t *src_ptr0, int8_t *src_ptr1, int8_t *dst_ptr, ES_S32 elem_size);

//  The prepare function on the kernel side, reserved as an extension interface
//  (common part, connected to runtime, function name and ARGs cannot be modified)
ES_S32 prepare(es_dsp_buffer_group *params)
{
    //  Retain the interface, currently no implementation is required
    return ES_DSP_SUCCESS;
}

//  The eval function on the kernel side allows users to fill in the required operator implementation
//  (common part, access runtime, function name and ARGs cannot be modified)
ES_S32 eval(es_dsp_buffer_group *params, es_dsp_buffer_group *inputs, es_dsp_buffer_group *outputs)
{
    //  Use TIME-STAMP to get program execution time
    ES_S32 eval_start, eval_stop;
#pragma no_reorder
    TIME_STAMP(eval_start);
#pragma no_reorder

    //  Step 1: Copy the operator config using dma to DRAM
    ADD_CONFIG_S *config_ddr = (ADD_CONFIG_S *)(params->buffers[0].addr);
    ADD_CONFIG_S *config_dram = (ADD_CONFIG_S *)(get_available_dram1_base() + (ES_U32)DRAM1_CONFIG_OFFSET);
    {
        const uint8_t CHANNEL = 0;

        uint32_t row_sz = sizeof(ADD_CONFIG_S);
        uint32_t nrows = 1;
        uint32_t ntiles = 1;
        uint32_t src_row_pitch = 1;
        uint32_t dst_row_pitch = 1;
        uint32_t src_tile_pitch = 1;
        uint32_t dst_tile_pitch = 1;

        //  Initiate the dma copy of data from DDR to DRAM
        idma_copy_3d_desc64(CHANNEL, &config_dram, &config_ddr, IDMA_DISABLE_INT, row_sz, nrows, ntiles,
                            src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch);
        //  Waiting for dma to finish copying data
        while (idma_hw_num_outstanding(CHANNEL) > 0) {
        }
    }

    //  Step 2: Get pointers to input/output data (at this time, the data is on the DDR)
    int8_t *ifmap0_ddr_ptr = (int8_t *)inputs->buffers[0].addr;
    int8_t *ifmap1_ddr_ptr = (int8_t *)inputs->buffers[1].addr;
    int8_t *ofmap0_ddr_ptr = (int8_t *)outputs->buffers[0].addr;

    //  Step 3: Obtain the starting addresses and buffer sizes of DRAM and spad
    //  spad_base_ptr: A DDR memory block requested by the framework, which can be used directly
    char *dram0_base_ptr = (char *)((ES_U32)get_available_dram0_base());
    char *dram1_base_ptr = (char *)((ES_U32)get_available_dram1_base());
    char *spad_base_ptr = (char *)((ES_U32)get_available_spad_base());

    ES_U32 dram0_size = get_available_dram0_size();
    ES_U32 dram1_size = get_available_dram1_size();
    ES_U32 spad_size = get_available_spad_size();

    //  Step 4: Allocate DRAM space for two inputs and one output. Assuming the data volume is small, DRAM can be placed directly without partitioning
    int8_t *ifmap0_dram_ptr = (int8_t *)dram0_base_ptr;
    int8_t *ifmap1_dram_ptr = (int8_t *)(dram0_base_ptr + config_dram->elem_size * sizeof(int8_t));
    int8_t *ofmap0_dram_ptr = (int8_t *)dram1_base_ptr;

    //  Step 5: Copy the two inputs to DRAM (to ensure concise and clear code, the sample example does not show how to ping pong)
    {
        const uint8_t CHANNEL = 1;

        uint32_t row_sz = config_dram->elem_size * sizeof(int8_t);
        uint32_t nrows = 1;
        uint32_t ntiles = 1;
        uint32_t src_row_pitch = 1;
        uint32_t dst_row_pitch = 1;
        uint32_t src_tile_pitch = 1;
        uint32_t dst_tile_pitch = 1;

        //  Copy ifmap0 from DDR to DRAM
        idma_copy_3d_desc64(CHANNEL, &ifmap0_dram_ptr, &ifmap0_ddr_ptr, IDMA_DISABLE_INT, row_sz, nrows, ntiles,
                            src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch);
        //  Copy ifmap1 from DDR to DRAM
        idma_copy_3d_desc64(CHANNEL, &ifmap1_dram_ptr, &ifmap1_ddr_ptr, IDMA_DISABLE_INT, row_sz, nrows, ntiles,
                            src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch);
        while (idma_hw_num_outstanding(CHANNEL) > 0) {
        }
    }

    // step 6: Start calculating ofmap0 = ifmap0 + ifmap1, where both input and output data are stored on DRAM
    add_kernel(ifmap0_dram_ptr, ifmap1_dram_ptr, ofmap0_dram_ptr, config_dram->elem_size);

    //  Step 7: Copy the output to DDR
    {
        const uint8_t CHANNEL = 2;

        uint32_t row_sz = config_dram->elem_size * sizeof(int8_t);
        uint32_t nrows = 1;
        uint32_t ntiles = 1;
        uint32_t src_row_pitch = 1;
        uint32_t dst_row_pitch = 1;
        uint32_t src_tile_pitch = 1;
        uint32_t dst_tile_pitch = 1;

        idma_copy_3d_desc64(CHANNEL, &ofmap0_ddr_ptr, &ofmap0_dram_ptr, IDMA_DISABLE_INT, row_sz, nrows, ntiles,
                            src_row_pitch, dst_row_pitch, src_tile_pitch, dst_tile_pitch);
        while (idma_hw_num_outstanding(CHANNEL) > 0) {
        }
    }

#pragma no_reorder
    TIME_STAMP(eval_stop);
#pragma no_reorder

    es_printf("dram0_size is %d, dram1_size is %d, spad_size is %d\n", dram0_size, dram1_size, spad_size);
    es_printf("this is eval of add op, using cycle is %d\n", eval_stop - eval_start);
    return ES_DSP_SUCCESS;
}

//  Specific implementation part of operator calculation function            (user-defined)
ES_S32 add_kernel(int8_t *ifmap0_ptr, int8_t *ifmap1_ptr, int8_t *ofmap0_ptr, ES_S32 elem_size)
{
    //  Declare the vector registers and aligned memory access registers used
    xb_vecNx16 vec_ifmap0, vec_ifmap1;
    xb_vecNx16 vec_ofmap0;
    valign va_ifmap0 = IVP_ZALIGN();
    valign va_ifmap1 = IVP_ZALIGN();
    valign va_dst = IVP_ZALIGN();

    //  Initialize vec_max and vectmin to the maximum and minimum values within the range of int8
    xb_vecNx16 vec_max = IVP_MOVVA16(127);
    xb_vecNx16 vec_min = IVP_MOVVA16(-128);

    //  Initialize input and output pointers, ready to start calculations
    xb_vecNx8 *__restrict ifmap0_vptr = (xb_vecNx8 *)(ifmap0_ptr);
    xb_vecNx8 *__restrict ifmap1_vptr = (xb_vecNx8 *)(ifmap1_ptr);
    xb_vecNx8 *__restrict ofmap0_vptr = (xb_vecNx8 *)(ofmap0_ptr);
    va_ifmap0 = IVP_LANX8S_PP(ifmap0_vptr);
    va_ifmap1 = IVP_LANX8S_PP(ifmap1_vptr);

    //  Start calculating
    for (ES_S32 x = 0; x < elem_size; x += XCHAL_IVPN_SIMD_WIDTH)
    {
        //  Load two input vector data into the register
        IVP_LANX8S_IP(vec_ifmap0, va_ifmap0, ifmap0_vptr);
        IVP_LANX8S_IP(vec_ifmap1, va_ifmap1, ifmap1_vptr);
        //  Add and clip to the range of int8
        vec_ofmap0 = IVP_ADDNX16(vec_ifmap0, vec_ifmap1);
        vec_ofmap0 = IVP_MAXNX16(IVP_MINNX16(vec_ofmap0, vec_max), vec_min);
        // Write the output to memory
        IVP_SAVNX8S_XP(vec_ofmap0, va_dst, ofmap0_vptr, XCHAL_IVPN_SIMD_WIDTH);
    }
    // Ensure that the result of flashing the register is saved to memory
    IVP_SAPOSNX8S_FP(va_dst, ofmap0_vptr);

    return 0;
}

//  Return pointers to the prepare and eval functions to the framework, non modifiable            (public section)
__attribute__((visibility("default"))) void *get_prepare(void) { return (void *)prepare; }
__attribute__((visibility("default"))) void *get_eval(void) { return (void *)eval; }

