// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include "common.h"
#include "es_dsp_ll_api.h"
#include "user_add.h"

static void printUsage()
{
    DSP_LOG_MSG("Usage: dsp_kernel_test -c <coreid> -i <input1.bin input2.bin ... inputN.bin> "
                "-o <output_dir> -l <log_level>\n");
}

// Users fill in ES_DSP_OPERATOR_DESC_S and their own operator config based on their own operator situation
int prepare_add_op_desc(ES_DSP_OPERATOR_DESC_S* op_desc_ptr, int32_t desc_size, int32_t add_elem_size)
{
    // Step 1: Fill the ES_DSP_OPERATOR_DESC_S structure
    /*
     * Step 1.1: Define the operator name. Note that this name should be consistent with the operator name on the
     * ernel side. The framework will load the corresponding operator library based on the operator name
     */
    memset(op_desc_ptr->operatorName, 0, KERNEL_LIB_NAME_MAXLEN);
    memcpy(op_desc_ptr->operatorName, "dsp_kernel_add", strlen("dsp_kernel_add") + 1);

    // Step 1.2: Fill in the buffer size described by the operator, including ES_DSP_OPERATOR_DESC_S and the size of the operator's own config
    op_desc_ptr->totalSize = desc_size;
    // Step 1.3: Fill in the number of operator config, excluding ES_DSP_OPERATOR_DESC_S. For example, this operator only has one ADD_CONFIG_S
    op_desc_ptr->bufferCntCfg = 1;
    // Step 1.4: Fill in the number of feature maps for the input operator, which is the add operator with two input tensors
    op_desc_ptr->bufferCntInput = 2;
    // Step 1.5: Fill the number of feature maps output by the operator, which is the add operator and has an output tensor
    op_desc_ptr->bufferCntOutput = 1;

    // Step 1.6: This operator has a total of config, two input tensors, and one output tensor (excluding the ES_DSP_OPERATOR_DESC_S structure).
    // Fill in the buffer size for each section in the order of config -> ifmap -> ofmap here
    op_desc_ptr->bufferSize[0] = sizeof(ADD_CONFIG_S);
    op_desc_ptr->bufferSize[1] = add_elem_size * sizeof(int8_t);
    op_desc_ptr->bufferSize[2] = add_elem_size * sizeof(int8_t);
    op_desc_ptr->bufferSize[3] = add_elem_size * sizeof(int8_t);

    // Step 2: Fill in the unique config of the operator itself, for example, the operator needs to fill ADD_CONFIG_S structure
    ADD_CONFIG_S* add_cfg = (ADD_CONFIG_S*)(op_desc_ptr + 1);
    add_cfg->elem_size = add_elem_size;

    return 0;
}

int main(int argc, char** argv)
{
    if (argc < 2) {
        printUsage();
        exit(-1);
    }

    ES_S32 status = ES_DSP_SUCCESS;
    ES_S32 ret;
    ES_S32 file_idx;
    ES_S32 dspFd;

    // Step0: Parse input arguments
    CMD_ARGS_S args;
    memset(&args, 0, sizeof(args));
    if (parseCmdArgs(argc, argv, &args) < 0) {
        printUsage();
        exit(-1);
    }
    DSP_LOG_MSG("CMD dspId:%d\n", args.dspId);
    DSP_LOG_MSG("CMD operatorDir:%s\n", args.operatorDir);
    DSP_LOG_MSG("CMD outputDir:%s\n", args.outputDir);
    DSP_LOG_MSG("CMD inputFileNum:%d\n", args.inputFileNum);
    for (file_idx = 0; file_idx < args.inputFileNum; file_idx++) {
        DSP_LOG_MSG("CMD input file:%s\n", args.inputFiles[file_idx]);
    }

    // Step1: Get version info
    ES_U64 version;
    ES_CHAR versionStr[VERSION_NAME_MAXLEN];
    ES_DSP_Capability_S capability;
    status = ES_DSP_LL_GetVersion(&version);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("get version failed\n");
        return status;
    }
    DSP_LOG_MSG("get version=%llu\n", version);

    status = ES_DSP_LL_GetVersionString(versionStr, VERSION_NAME_MAXLEN);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("get version string failed\n");
        return status;
    }
    DSP_LOG_MSG("get version string=%s\n", versionStr);

    status = ES_DSP_LL_GetCapability(&capability);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("get version string failed\n");
        return status;
    }
    DSP_LOG_MSG("get capability=%llu\n", capability.reserved);

    status = ES_DSP_LL_SetLogLevel(args.logLevel);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("set log level failed\n");
        return status;
    }
    DSP_LOG_MSG("set log level=%d\n", args.logLevel);

    // Step2: Open DSP device
    status = ES_DSP_LL_Open(args.dspId, &dspFd);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("open dsp%d failed\n", args.dspId);
        return status;
    }

    // Step3: Prepare DSP operator description and input data.
    ES_DSP_OPERATOR_DESC_S* opDesc;
    ES_CHAR descVec[1024];
    opDesc = (ES_DSP_OPERATOR_DESC_S*)descVec;
    // This sample example is: adding two arrays of int8 with a length of 1024 to obtain an output of int8 with a length of 1024. The data is clipped to [-128,127]
    int32_t add_elem_size = 1024;
    int32_t desc_size = sizeof(ES_DSP_OPERATOR_DESC_S) + sizeof(ADD_CONFIG_S);
    prepare_add_op_desc(opDesc, desc_size, add_elem_size);

    ES_DEVICE_BUFFER_GROUP_S params;
    ES_DEVICE_BUFFER_GROUP_S inputs;
    ES_DEVICE_BUFFER_GROUP_S outputs;
    ret = loadParamsData(opDesc, &params);
    if (ret != 0) {
        DSP_LOG_ERR("loadParamsData failed.\n");
        return ret;
    }
    prepareDmaBuffers(dspFd, &params, opDesc->bufferCntCfg);
    ret = loadInputsData(args.inputFiles, opDesc, &inputs);
    if (ret != 0) {
        DSP_LOG_ERR("loadInputsData failed.\n");
        return ret;
    }
    prepareDmaBuffers(dspFd, &inputs, opDesc->bufferCntInput);
    ret = prepareOutputBuffer(opDesc, &outputs);
    if (ret != 0) {
        DSP_LOG_ERR("prepareOutputBuffer failed.\n");
        return ret;
    }
    prepareDmaBuffers(dspFd, &outputs, opDesc->bufferCntOutput);
    DSP_LOG_MSG("Step3: Load data from for DSP operator %s done\n", opDesc->operatorName);

    // Step4: Load DSP operator
    ES_DSP_HANDLE handle;
    status = ES_DSP_LL_LoadOperator(dspFd, opDesc->operatorName, args.operatorDir, &handle);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("load operator %s failed\n", opDesc->operatorName);
        return status;
    }
    DSP_LOG_MSG("Step4: Load operator %s done\n", opDesc->operatorName);

    // Step5: Execute DSP operator task with sync mode.
    ES_DSP_TASK_S task;
    memset(&task, 0, sizeof(ES_DSP_TASK_S));
    task.operatorHandle = handle;
    prepareTaskBuffer(&task, &params, &inputs, &outputs);

    status = ES_DSP_LL_SubmitTask(dspFd, &task);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("submit %s task(op_handle=0x%x sync=0x%x mode=0x%x) failed\n", opDesc->operatorName,
                    task.operatorHandle, task.syncCache, task.pollMode);
        return status;
    }
    DSP_LOG_MSG("Step5: Execute operator %s done\n", opDesc->operatorName);

    // Step6: Unload DSP operator
    status = ES_DSP_LL_UnloadOperator(dspFd, handle);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("unload operator %s failed\n", opDesc->operatorName);
        return status;
    }
    DSP_LOG_MSG("Step6: Unload operator %s done\n", opDesc->operatorName);

    // Step7: Get DSP operator task results
    if (!args.outputDir) {
        args.outputDir = dirname(args.inputFiles[0]);
    }
    status = dumpDspOutput(&outputs, args.outputDir);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("dump output failed.\n");
    }

    unPrepareDmaBuffers(dspFd, &params, opDesc->bufferCntCfg);
    unPrepareDmaBuffers(dspFd, &inputs, opDesc->bufferCntInput);
    unPrepareDmaBuffers(dspFd, &outputs, opDesc->bufferCntOutput);
    freeDeviceBuffers(&params, &inputs, &outputs);
    DSP_LOG_MSG("Step7: Dump DSP operator task results done.\n");

    // Step8: Close DSP device
    status = ES_DSP_LL_Close(dspFd);
    if (status != ES_DSP_SUCCESS) {
        DSP_LOG_ERR("release device %d failed\n", args.dspId);
        return status;
    }
    DSP_LOG_MSG("Step8: Close DSP device %d done.\n", args.dspId);

    return 0;
}
